@@ -19,32 +19,31 @@
 """
 System-level utilities and helper functions.
 """
-
-import errno
-
 try:
     from eventlet import sleep
 except ImportError:
     from time import sleep
 from eventlet.green import socket
 
-import functools
 import hashlib
 import os
 import re
-import uuid
 
+import glance_store
 from OpenSSL import crypto
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_utils import encodeutils
 from oslo_utils import excutils
 from oslo_utils import timeutils
+from oslo_utils import uuidutils
+from oslo_versionedobjects import fields
+import requests
 import six
-from webob import exc
 
 from glare.common import exception
-from glare.i18n import _, _LE, _LW
+from glare.i18n import _
+from glare.objects.meta import fields as glare_fields
 
 CONF = cfg.CONF
 
@@ -53,35 +52,8 @@ LOG = logging.getLogger(__name__)
 GLARE_TEST_SOCKET_FD_STR = 'GLARE_TEST_SOCKET_FD'
 
 
-def chunkreadable(iter, chunk_size=65536):
-    """
-    Wrap a readable iterator with a reader yielding chunks of
-    a preferred size, otherwise leave iterator unchanged.
-
-    :param iter: an iter which may also be readable
-    :param chunk_size: maximum size of chunk
-    """
-    return chunkiter(iter, chunk_size) if hasattr(iter, 'read') else iter
-
-
-def chunkiter(fp, chunk_size=65536):
-    """
-    Return an iterator to a file-like obj which yields fixed size chunks
-
-    :param fp: a file-like object
-    :param chunk_size: maximum size of chunk
-    """
-    while True:
-        chunk = fp.read(chunk_size)
-        if chunk:
-            yield chunk
-        else:
-            break
-
-
 def cooperative_iter(iter):
-    """
-    Return an iterator which schedules after each
+    """Return an iterator which schedules after each
     iteration. This can prevent eventlet thread starvation.
 
     :param iter: an iterator to wrap
@@ -92,13 +64,11 @@ def cooperative_iter(iter):
             yield chunk
     except Exception as err:
         with excutils.save_and_reraise_exception():
-            msg = _LE("Error: cooperative_iter exception %s") % err
-            LOG.error(msg)
+            LOG.error("Error: cooperative_iter exception %s", err)
 
 
 def cooperative_read(fd):
-    """
-    Wrap a file descriptor's read with a partial function which schedules
+    """Wrap a file descriptor's read with a partial function which schedules
     after each read. This can prevent eventlet thread starvation.
 
     :param fd: a file descriptor to wrap
@@ -114,8 +84,7 @@ MAX_COOP_READER_BUFFER_SIZE = 134217728  # 128M seems like a sane buffer limit
 
 
 class CooperativeReader(object):
-    """
-    An eventlet thread friendly class for reading in blob data.
+    """An eventlet thread friendly class for reading in blob data.
 
     When accessing data either through the iterator or the read method
     we perform a sleep to allow a co-operative yield. When there is more than
@@ -124,8 +93,7 @@ class CooperativeReader(object):
     having the same thread be continuously active.
     """
     def __init__(self, fd):
-        """
-        :param fd: Underlying blob file object
+        """:param fd: Underlying blob file object
         """
         self.fd = fd
         self.iterator = None
@@ -183,7 +151,7 @@ class CooperativeReader(object):
                     # check should never fire. Regardless it still worths to
                     # make the check, as the code may be reused somewhere else.
                     if len(result) >= MAX_COOP_READER_BUFFER_SIZE:
-                        raise exception.LimitExceeded()
+                        raise exception.RequestEntityTooLarge()
                     self.position += len(chunk)
                 else:
                     try:
@@ -202,8 +170,7 @@ class CooperativeReader(object):
 
 
 class LimitingReader(object):
-    """
-    Reader designed to fail when reading blob data past the configured
+    """Reader designed to fail when reading blob data past the configured
     allowable amount.
     """
     def __init__(self, data, limit):
@@ -214,6 +181,7 @@ class LimitingReader(object):
         self.data = data
         self.limit = limit
         self.bytes_read = 0
+        self.md5 = hashlib.md5()
         self.sha1 = hashlib.sha1()
         self.sha256 = hashlib.sha256()
 
@@ -225,59 +193,20 @@ class LimitingReader(object):
             else:
                 yield chunk
 
-    def read(self, i):
-        result = self.data.read(i)
-        len_result = len(result)
+    def read(self, length=None):
+        res = self.data.read() if length is None else self.data.read(length)
+        len_result = len(res)
         self.bytes_read += len_result
         if len_result:
-            self.sha1.update(result)
-            self.sha256.update(result)
+            self.md5.update(res)
+            self.sha1.update(res)
+            self.sha256.update(res)
         if self.bytes_read > self.limit:
-            raise exception.RequestEntityTooLarge()
-        return result
-
-
-def safe_mkdirs(path):
-    try:
-        os.makedirs(path)
-    except OSError as e:
-        if e.errno != errno.EEXIST:
-            raise
-
-
-def mutating(func):
-    """Decorator to enforce read-only logic"""
-    @functools.wraps(func)
-    def wrapped(self, req, *args, **kwargs):
-        if req.context.read_only:
-            msg = "Read-only access"
-            LOG.debug(msg)
-            raise exc.HTTPForbidden(msg, request=req,
-                                    content_type="text/plain")
-        return func(self, req, *args, **kwargs)
-    return wrapped
-
-
-def setup_remote_pydev_debug(host, port):
-    error_msg = _LE('Error setting up the debug environment. Verify that the'
-                    ' option pydev_worker_debug_host is pointing to a valid '
-                    'hostname or IP on which a pydev server is listening on'
-                    ' the port indicated by pydev_worker_debug_port.')
-
-    try:
-        try:
-            from pydev import pydevd
-        except ImportError:
-            import pydevd
-
-        pydevd.settrace(host,
-                        port=port,
-                        stdoutToServer=True,
-                        stderrToServer=True)
-        return True
-    except Exception:
-        with excutils.save_and_reraise_exception():
-            LOG.exception(error_msg)
+            message = _("The server is refusing to process a request because"
+                        " the request entity is larger than the server is"
+                        " willing or able to process - %s bytes.") % self.limit
+            raise exception.RequestEntityTooLarge(message=message)
+        return res
 
 
 def validate_key_cert(key_file, cert_file):
@@ -309,21 +238,21 @@ def validate_key_cert(key_file, cert_file):
                             'ce': ce})
 
     try:
-        data = str(uuid.uuid4())
+        data = uuidutils.generate_uuid()
         # On Python 3, explicitly encode to UTF-8 to call crypto.sign() which
         # requires bytes. Otherwise, it raises a deprecation warning (and
         # will raise an error later).
         data = encodeutils.to_utf8(data)
         digest = CONF.digest_algorithm
         if digest == 'sha1':
-            LOG.warn(
-                _LW('The FIPS (FEDERAL INFORMATION PROCESSING STANDARDS)'
-                    ' state that the SHA-1 is not suitable for'
-                    ' general-purpose digital signature applications (as'
-                    ' specified in FIPS 186-3) that require 112 bits of'
-                    ' security. The default value is sha1 in Kilo for a'
-                    ' smooth upgrade process, and it will be updated'
-                    ' with sha256 in next release(L).'))
+            LOG.warning(
+                'The FIPS (FEDERAL INFORMATION PROCESSING STANDARDS)'
+                ' state that the SHA-1 is not suitable for'
+                ' general-purpose digital signature applications (as'
+                ' specified in FIPS 186-3) that require 112 bits of'
+                ' security. The default value is sha1 in Kilo for a'
+                ' smooth upgrade process, and it will be updated'
+                ' with sha256 in next release(L).')
         out = crypto.sign(key, data, digest)
         crypto.verify(cert, out, data, digest)
     except crypto.Error as ce:
@@ -357,9 +286,8 @@ except re.error:
 
 
 def no_4byte_params(f):
-    """
-    Checks that no 4 byte unicode characters are allowed
-    in dicts' keys/values and string's parameters
+    """Checks that no 4 byte unicode characters are allowed
+    in dicts' keys/values and string's parameters.
     """
     def wrapper(*args, **kwargs):
 
@@ -369,17 +297,17 @@ def no_4byte_params(f):
 
         def _check_dict(data_dict):
             # a dict of dicts has to be checked recursively
-            for key, value in six.iteritems(data_dict):
+            for key, value in data_dict.items():
                 if isinstance(value, dict):
                     _check_dict(value)
                 else:
                     if _is_match(key):
                         msg = _("Property names can't contain 4 byte unicode.")
-                        raise exception.Invalid(msg)
+                        raise exception.BadRequest(msg)
                     if _is_match(value):
                         msg = (_("%s can't contain 4 byte unicode characters.")
                                % key.title())
-                        raise exception.Invalid(msg)
+                        raise exception.BadRequest(msg)
 
         for data_dict in [arg for arg in args if isinstance(arg, dict)]:
             _check_dict(data_dict)
@@ -387,7 +315,7 @@ def no_4byte_params(f):
         for arg in args:
             if _is_match(arg):
                 msg = _("Param values can't contain 4 byte unicode.")
-                raise exception.Invalid(msg)
+                raise exception.BadRequest(msg)
         # check kwargs as well, as params are passed as kwargs via
         # registry calls
         _check_dict(kwargs)
@@ -396,8 +324,7 @@ def no_4byte_params(f):
 
 
 def stash_conf_values():
-    """
-    Make a copy of some of the current global CONF's settings.
+    """Make a copy of some of the current global CONF's settings.
     Allows determining if any of these values have changed
     when the config is reloaded.
     """
@@ -407,7 +334,9 @@ def stash_conf_values():
         'tcp_keepidle': CONF.cert_file,
         'backlog': CONF.backlog,
         'key_file': CONF.key_file,
-        'cert_file': CONF.cert_file
+        'cert_file': CONF.cert_file,
+        'enabled_artifact_types': CONF.enabled_artifact_types,
+        'custom_artifact_types_modules': CONF.custom_artifact_types_modules
     }
 
     return conf
@@ -419,8 +348,7 @@ def split_filter_op(expression):
     When no operator is found, default to an equality comparison.
 
     :param expression: the expression to parse
-
-    :returns: a tuple (operator, threshold) parsed from expression
+    :return: a tuple (operator, threshold) parsed from expression
     """
     left, sep, right = expression.partition(':')
     if sep:
@@ -489,53 +417,27 @@ def split_filter_value_for_quotes(value):
     return [val[0] or val[1] for val in re.findall(tmp, value)]
 
 
-def evaluate_filter_op(value, operator, threshold):
-    """Evaluate a comparison operator.
-    Designed for use on a comparative-filtering query field.
-
-    :param value: evaluated against the operator, as left side of expression
-    :param operator: any supported filter operation
-    :param threshold: to compare value against, as right side of expression
-
-    :raises: InvalidFilterOperatorValue if an unknown operator is provided
-
-    :returns: boolean result of applied comparison
-
-    """
-    if operator == 'gt':
-        return value > threshold
-    elif operator == 'gte':
-        return value >= threshold
-    elif operator == 'lt':
-        return value < threshold
-    elif operator == 'lte':
-        return value <= threshold
-    elif operator == 'neq':
-        return value != threshold
-    elif operator == 'eq':
-        return value == threshold
-
-    msg = _("Unable to filter on a unknown operator.")
-    raise exception.InvalidFilterOperatorValue(msg)
-
-
 class error_handler(object):
     def __init__(self, error_map, default_exception=None):
+        """Init method of the class.
+
+        :param error_map: dict of exception that can be raised
+        in func and exceptions that must be raised for these exceptions.
+        For example, if sqlalchemy NotFound might be raised and we need
+        re-raise it as glare NotFound exception then error_map must
+        contain {"catch": SQLAlchemyNotFound,
+                 "raise": exceptions.NotFound}
+        :param default_exception: default exception that must be raised if
+        exception that cannot be found in error map was raised
+        :return: func
+        """
         self.error_map = error_map
         self.default_exception = default_exception
 
     def __call__(self, f):
-        """Decorator that catches exception that came from func or method
-            :param f: targer func
-            :param error_map: dict of exception that can be raised
-            in func and exceptions that must be raised for these exceptions.
-            For example, if sqlalchemy NotFound might be raised and we need
-            re-raise it as glare NotFound exception then error_map must
-            contain {"catch": SQLAlchemyNotFound,
-                     "raise": exceptions.NotFound}
-            :param default_exception: default exception that must be raised if
-            exception that cannot be found in error map was raised
-            :return: func
+        """Decorator that catches exception that came from func or method.
+
+        :param f: target func
         """
 
         def new_function(*args, **kwargs):
@@ -553,40 +455,171 @@ class error_handler(object):
         return new_function
 
 
-class DictDiffer(object):
-    """
-    Calculate the difference between two dictionaries as:
-    (1) items added
-    (2) items removed
-    (3) keys same in both but changed values
-    (4) keys same in both and unchanged values
+def get_schema_type(attr):
+    if isinstance(attr, fields.IntegerField) or attr is fields.Integer:
+        return 'integer'
+    elif isinstance(attr, fields.FloatField) or attr is fields.Float:
+        return 'number'
+    elif isinstance(attr, fields.FlexibleBooleanField) \
+            or attr is fields.FlexibleBoolean:
+        return 'boolean'
+    elif isinstance(attr, glare_fields.List):
+        return 'array'
+    elif isinstance(attr, (glare_fields.Dict, glare_fields.BlobField)):
+        return 'object'
+    return 'string'
+
+
+def get_glare_type(attr):
+    if isinstance(attr, fields.IntegerField):
+        return 'Integer'
+    elif isinstance(attr, fields.FloatField):
+        return 'Float'
+    elif isinstance(attr, fields.FlexibleBooleanField):
+        return 'Boolean'
+    elif isinstance(attr, fields.DateTimeField):
+        return 'DateTime'
+    elif isinstance(attr, glare_fields.BlobField):
+        return 'Blob'
+    elif isinstance(attr, glare_fields.Link):
+        return 'Link'
+    elif isinstance(attr, glare_fields.List):
+        return _get_element_type(attr.element_type) + 'List'
+    elif isinstance(attr, glare_fields.Dict):
+        return _get_element_type(attr.element_type) + 'Dict'
+    return 'String'
+
+
+def _get_element_type(element_type):
+    if element_type is fields.FlexibleBooleanField:
+        return 'Boolean'
+    elif element_type is fields.Integer:
+        return 'Integer'
+    elif element_type is fields.Float:
+        return 'Float'
+    elif element_type is glare_fields.BlobFieldType:
+        return 'Blob'
+    elif element_type is glare_fields.LinkFieldType:
+        return 'Link'
+    return 'String'
+
+
+class BlobIterator(object):
+    """Reads data from a blob, one chunk at a time.
     """
 
-    def __init__(self, current_dict, past_dict):
-        self.current_dict, self.past_dict = current_dict, past_dict
-        self.current_keys, self.past_keys = [
-            set(d.keys()) for d in (current_dict, past_dict)]
-        self.intersect = self.current_keys.intersection(
-            self.past_keys)
+    def __init__(self, data, chunk_size=65536):
+        self.chunk_size = chunk_size
+        self.data = data
 
-    def added(self):
-        return self.current_keys - self.intersect
+    def __iter__(self):
+        bytes_left = len(self.data)
+        i = 0
+        while bytes_left > 0:
+            data = self.data[i * self.chunk_size:(i + 1) * self.chunk_size]
+            bytes_left -= len(data)
+            yield data
+        raise StopIteration()
 
-    def removed(self):
-        return self.past_keys - self.intersect
 
-    def changed(self):
-        return set(o for o in self.intersect
-                   if self.past_dict[o] != self.current_dict[o])
+def validate_status_transition(af, from_status, to_status):
+    if from_status == 'deleted':
+        msg = _("Cannot change status if artifact is deleted.")
+        raise exception.Forbidden(msg)
+    if to_status == 'active':
+        if from_status == 'drafted':
+            for name, type_obj in af.fields.items():
+                if type_obj.required_on_activate and getattr(af, name) is None:
+                    msg = _("'%s' field value must be set before "
+                            "activation.") % name
+                    raise exception.Forbidden(msg)
+    elif to_status == 'drafted':
+        if from_status != 'drafted':
+            msg = _("Cannot change status to 'drafted'") % from_status
+            raise exception.Forbidden(msg)
+    elif to_status == 'deactivated':
+        if from_status not in ('active', 'deactivated'):
+            msg = _("Cannot deactivate artifact if it's not active.")
+            raise exception.Forbidden(msg)
+    elif to_status == 'deleted':
+        msg = _("Cannot delete artifact with PATCH requests. Use special "
+                "API to do this.")
+        raise exception.Forbidden(msg)
+    else:
+        msg = _("Unknown artifact status: %s.") % to_status
+        raise exception.BadRequest(msg)
 
-    def unchanged(self):
-        return set(o for o in self.intersect
-                   if self.past_dict[o] == self.current_dict[o])
 
-    def __str__(self):
-        msg = "\nResult output:\n"
-        msg += "\tAdded keys: %s\n" % ', '.join(self.added())
-        msg += "\tRemoved keys: %s\n" % ', '.join(self.removed())
-        msg += "\tChanged keys: %s\n" % ', '.join(self.changed())
-        msg += "\tUnchanged keys: %s\n" % ', '.join(self.unchanged())
-        return msg
+def validate_visibility_transition(af, from_visibility, to_visibility):
+    if to_visibility == 'private':
+        if from_visibility != 'private':
+            msg = _("Cannot make artifact private again.")
+            raise exception.Forbidden()
+    elif to_visibility == 'public':
+        if af.status != 'active':
+            msg = _("Cannot change visibility to 'public' if artifact"
+                    " is not active.")
+            raise exception.Forbidden(msg)
+    else:
+        msg = _("Unknown artifact visibility: %s.") % to_visibility
+        raise exception.BadRequest(msg)
+
+
+def validate_change_allowed(af, field_name):
+    """Validate if fields can be set for the artifact."""
+    if field_name not in af.fields:
+        msg = _("Cannot add new field '%s' to artifact.") % field_name
+        raise exception.BadRequest(msg)
+    if af.status not in ('active', 'drafted'):
+        msg = _("Forbidden to change fields "
+                "if artifact is not active or drafted.")
+        raise exception.Forbidden(message=msg)
+    if af.fields[field_name].system is True:
+        msg = _("Forbidden to specify system field %s. It is not "
+                "available for modifying by users.") % field_name
+        raise exception.Forbidden(msg)
+    if af.status == 'active' and not af.fields[field_name].mutable:
+        msg = (_("Forbidden to change field '%s' after activation.")
+               % field_name)
+        raise exception.Forbidden(message=msg)
+
+
+def initialize_glance_store():
+    """Initialize glance store."""
+    glance_store.register_opts(CONF)
+    set_glance_store_config_defaults()
+    glance_store.create_stores(CONF)
+    glance_store.verify_default_store()
+
+
+def set_glance_store_config_defaults():
+    # By default glance and glare share common place to store data.
+    # To prevent possible collisions we have to set other glance_store default
+    # values for various backends.
+    cfg.set_defaults(glance_store.backend._list_opts()[1][1],
+                     # default '/var/lib/glance/images
+                     filesystem_store_datadir='/var/lib/glare/artifacts',
+                     # default 'images'
+                     rbd_store_pool='artifacts',
+                     # default '/openstack_glance
+                     vmware_store_image_dir='/openstack_glare',
+                     # default 'glance'
+                     swift_store_container='glare')
+
+
+def get_system_ca_file():
+    """Return path to system default CA file."""
+    # Standard CA file locations for Debian/Ubuntu, RedHat/Fedora,
+    # Suse, FreeBSD/OpenBSD, MacOSX, and the bundled ca
+    ca_path = ['/etc/ssl/certs/ca-certificates.crt',
+               '/etc/pki/tls/certs/ca-bundle.crt',
+               '/etc/ssl/ca-bundle.pem',
+               '/etc/ssl/cert.pem',
+               '/System/Library/OpenSSL/certs/cacert.pem',
+               requests.certs.where()]
+    for ca in ca_path:
+        LOG.debug("Looking for ca file %s", ca)
+        if os.path.exists(ca):
+            LOG.debug("Using ca file %s", ca)
+            return ca
+    LOG.warning("System ca file could not be found.")
