@@ -12,12 +12,13 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import hashlib
 import operator
 import threading
-import uuid
 
 from oslo_config import cfg
 from oslo_db import exception as db_exception
+from oslo_db import options
 from oslo_db.sqlalchemy import session
 from oslo_log import log as os_logging
 from oslo_utils import timeutils
@@ -27,6 +28,7 @@ import six
 import sqlalchemy
 from sqlalchemy import and_
 import sqlalchemy.exc
+from sqlalchemy import exists
 from sqlalchemy import func
 from sqlalchemy import or_
 import sqlalchemy.orm as orm
@@ -37,20 +39,19 @@ from glare.common import exception
 from glare.common import semver_db
 from glare.common import utils
 from glare.db.sqlalchemy import models
-from glare.i18n import _, _LW
+from glare.i18n import _
 
 LOG = os_logging.getLogger(__name__)
 
 CONF = cfg.CONF
 CONF.import_group("profiler", "glare.common.wsgi")
+options.set_defaults(CONF)
 
 
 BASE_ARTIFACT_PROPERTIES = ('id', 'visibility', 'created_at', 'updated_at',
                             'activated_at', 'owner', 'status', 'description',
                             'name', 'type_name', 'version')
 
-DEFAULT_SORT_PARAMETERS = (('created_at', 'desc', None), ('id', 'asc', None))
-
 _FACADE = None
 _LOCK = threading.Lock()
 
@@ -59,7 +60,7 @@ def _retry_on_deadlock(exc):
     """Decorator to retry a DB API call if Deadlock was received."""
 
     if isinstance(exc, db_exception.DBDeadlock):
-        LOG.warn(_LW("Deadlock detected. Retrying..."))
+        LOG.warning("Deadlock detected. Retrying...")
         return True
     return False
 
@@ -89,35 +90,25 @@ def get_session(autocommit=True, expire_on_commit=False):
                               expire_on_commit=expire_on_commit)
 
 
-def clear_db_env():
-    """
-    Unset global configuration variables for database.
-    """
-    global _FACADE
-    _FACADE = None
+def setup_db():
+    engine = get_engine()
+    models.register_models(engine)
 
 
-def create(context, values, session):
-    return _create_or_update(context, None, values, session)
-
-
-def update(context, artifact_id, values, session):
-    return _create_or_update(context, artifact_id, values, session)
+def drop_db():
+    engine = get_engine()
+    models.unregister_models(engine)
 
 
 @retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
        stop_max_attempt_number=50)
 def delete(context, artifact_id, session):
-    artifact = _get(context, artifact_id, session)
-    artifact.properties = []
-    artifact.tags = []
-    artifact.status = 'deleted'
-    artifact.save(session=session)
+    with session.begin():
+        session.query(models.Artifact).filter_by(id=artifact_id).delete()
 
 
 def _drop_protected_attrs(model_class, values):
-    """
-    Removed protected attributes from values dictionary using the models
+    """Removed protected attributes from values dictionary using the models
     __protected_attributes__ field.
     """
     for attr in model_class.__protected_attributes__:
@@ -127,20 +118,14 @@ def _drop_protected_attrs(model_class, values):
 
 @retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
        stop_max_attempt_number=50)
-def _create_or_update(context, artifact_id, values, session):
+@utils.no_4byte_params
+def create_or_update(context, artifact_id, values, session):
     with session.begin():
         _drop_protected_attrs(models.Artifact, values)
         if artifact_id is None:
-            if 'type_name' not in values:
-                msg = _('Type name must be set.')
-                raise exception.BadRequest(msg)
             # create new artifact
             artifact = models.Artifact()
-            if 'id' not in values:
-                artifact.id = str(uuid.uuid4())
-            else:
-                artifact.id = values.pop('id')
-            artifact.created_at = timeutils.utcnow()
+            artifact.id = values.pop('id')
         else:
             # update the existing artifact
             artifact = _get(context, artifact_id, session)
@@ -161,10 +146,20 @@ def _create_or_update(context, artifact_id, values, session):
             artifact.blobs = _do_blobs(artifact, blobs)
 
         artifact.updated_at = timeutils.utcnow()
-        if 'status' in values and values['status'] == 'active':
-            artifact.activated_at = timeutils.utcnow()
+        if 'status' in values:
+            if session.query(exists().where(and_(
+                models.ArtifactBlob.status == 'saving',
+                models.ArtifactBlob.artifact_id == artifact_id))
+            ).one()[0]:
+                raise exception.Conflict(
+                    "You cannot change artifact status if it has "
+                    "uploading blobs.")
+            if values['status'] == 'active':
+                artifact.activated_at = timeutils.utcnow()
         artifact.update(values)
+
         artifact.save(session=session)
+        LOG.debug("Response from the database was received.")
 
         return artifact.to_dict()
 
@@ -176,7 +171,7 @@ def _get(context, artifact_id, session):
         artifact = query.one()
     except orm.exc.NoResultFound:
         msg = _("Artifact with id=%s not found.") % artifact_id
-        LOG.warn(msg)
+        LOG.warning(msg)
         raise exception.ArtifactNotFound(msg)
     return artifact
 
@@ -188,15 +183,16 @@ def get(context, artifact_id, session):
 def get_all(context, session, filters=None, marker=None, limit=None,
             sort=None, latest=False):
     """List all visible artifacts
+
     :param filters: dict of filter keys and values.
     :param marker: artifact id after which to start page
     :param limit: maximum number of artifacts to return
     :param sort: a tuple (key, dir, type) where key is an attribute by
-    which results should be sorted, dir is a direction: 'asc' or 'desc',
-    and type is type of the attribute: 'bool', 'string', 'numeric' or 'int' or
-    None if attribute is base.
+     which results should be sorted, dir is a direction: 'asc' or 'desc',
+     and type is type of the attribute: 'bool', 'string', 'numeric' or 'int' or
+     None if attribute is base.
     :param latest: flag that indicates, that only artifacts with highest
-    versions should be returned in output
+     versions should be returned in output
     """
     artifacts = _get_all(
         context, session, filters, marker, limit, sort, latest)
@@ -280,13 +276,6 @@ def _get_all(context, session, filters=None, marker=None, limit=None,
     if marker is not None:
         marker_artifact = get(context, marker, session)
 
-    if sort is None:
-        sort = DEFAULT_SORT_PARAMETERS
-    else:
-        for val in DEFAULT_SORT_PARAMETERS:
-            if val not in sort:
-                sort.append(val)
-
     query = _do_paginate_query(query=query, limit=limit,
                                marker=marker_artifact, sort=sort)
 
@@ -379,7 +368,7 @@ def _do_paginate_query(query, marker=None, limit=None, sort=None):
     return query
 
 
-def _do_artifacts_query(context, session, latest=False):
+def _do_artifacts_query(context, session):
     """Build the query to get all artifacts based on the context"""
 
     query = session.query(models.Artifact)
@@ -392,9 +381,6 @@ def _do_artifacts_query(context, session, latest=False):
 
 
 def _apply_query_base_filters(query, context):
-    # Don't show deleted artifacts
-    query = query.filter(models.Artifact.status != 'deleted')
-
     # If admin, return everything.
     if context.is_admin:
         return query
@@ -444,12 +430,21 @@ def _do_query_filters(filters):
             else:
                 if field_name == 'version':
                     value = [semver_db.parse(val) for val in value]
-                basic_conds.append(
-                    [getattr(models.Artifact, field_name).in_(value)])
+                    basic_conds.append(
+                        [or_(*[
+                            models.Artifact.version == ver for ver in value])])
+                else:
+                    basic_conds.append(
+                        [getattr(models.Artifact, field_name).in_(value)])
         else:
             conds = [models.ArtifactProperty.name == field_name]
             if key_name is not None:
-                conds.extend([models.ArtifactProperty.key_name == key_name])
+                if op == 'eq' or value is not None:
+                    conds.extend(
+                        [models.ArtifactProperty.key_name == key_name])
+                elif op == 'in':
+                    conds.extend(
+                        [models.ArtifactProperty.key_name.in_(key_name)])
             if value is not None:
                 if op != 'in':
                     fn = op_mappings[op]
@@ -506,7 +501,7 @@ def _do_properties(artifact, new_properties):
         if prop.name not in new_properties:
             props_to_update.append(prop)
 
-    for prop_name, prop_value in six.iteritems(new_properties):
+    for prop_name, prop_value in new_properties.items():
         if prop_value is None:
             continue
         if isinstance(prop_value, list):
@@ -524,7 +519,7 @@ def _do_properties(artifact, new_properties):
                         _create_property(prop_name, list_prop, position=pos)
                     )
         elif isinstance(prop_value, dict):
-            for dict_key, dict_val in six.iteritems(prop_value):
+            for dict_key, dict_val in prop_value.items():
                 for prop in artifact.properties:
                     if prop.name == prop_name and prop.key_name == dict_key:
                         if getattr(prop, _get_prop_type(dict_val)) != dict_val:
@@ -563,7 +558,7 @@ def _do_blobs(artifact, new_blobs):
         if blob.name not in new_blobs:
             blobs_to_update.append(blob)
 
-    for blob_name, blob_value in six.iteritems(new_blobs):
+    for blob_name, blob_value in new_blobs.items():
         if blob_value is None:
             continue
         if isinstance(blob_value.get('status'), str):
@@ -578,7 +573,7 @@ def _do_blobs(artifact, new_blobs):
                 _update_blob_values(blob, blob_value)
                 blobs_to_update.append(blob)
         else:
-            for dict_key, dict_val in six.iteritems(blob_value):
+            for dict_key, dict_val in blob_value.items():
                 for blob in artifact.blobs:
                     if blob.name == blob_name and blob.key_name == dict_key:
                         _update_blob_values(blob, dict_val)
@@ -594,26 +589,164 @@ def _do_blobs(artifact, new_blobs):
     return blobs_to_update
 
 
+def count_artifact_number(context, session, type_name=None):
+    """Return a number of artifacts for tenant."""
+    query = session.query(func.count(models.Artifact.id)).filter(
+        models.Artifact.owner == context.tenant)
+    if type_name is not None:
+        query = query.filter(models.Artifact.type_name == type_name)
+    return query.order_by(None).scalar() or 0
+
+
+def calculate_uploaded_data(context, session, type_name=None):
+    """Return the amount of uploaded data for tenant."""
+    query = session.query(
+        func.sum(models.ArtifactBlob.size)).join(
+        models.Artifact, aliased=True).filter(
+        models.Artifact.owner == context.tenant)
+    if type_name is not None:
+        query = query.filter(models.Artifact.type_name == type_name)
+    return query.order_by(None).scalar() or 0
+
+
+def _generate_quota_id(project_id, quota_name, type_name=None):
+    quota_id = b"%s:%s" % (project_id.encode(), quota_name.encode())
+    if type_name is not None:
+        quota_id += b":%s" % type_name.encode()
+    return hashlib.md5(quota_id).hexdigest()
+
+
 @retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
        stop_max_attempt_number=50)
+@utils.no_4byte_params
+def set_quotas(values, session):
+    """Create new quota instances in database"""
+    with session.begin():
+        for project_id, project_quotas in values.items():
+
+            # reset all project quotas
+            session.query(models.ArtifactQuota).filter(
+                models.ArtifactQuota.project_id == project_id).delete()
+
+            # generate new quotas
+            for quota_name, quota_value in project_quotas.items():
+                q = models.ArtifactQuota()
+                q.project_id = project_id
+                q.quota_name = quota_name
+                q.quota_value = quota_value
+                session.add(q)
+
+        # save all quotas
+        session.flush()
+
+
+@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
+       stop_max_attempt_number=50)
+def get_all_quotas(session, project_id=None):
+    """List all available quotas."""
+    query = session.query(models.ArtifactQuota)
+    if project_id is not None:
+        query = query.filter(
+            models.ArtifactQuota.project_id == project_id)
+    quotas = query.order_by(models.ArtifactQuota.project_id).all()
+
+    res = {}
+    for quota in quotas:
+        res.setdefault(
+            quota.project_id, {})[quota.quota_name] = quota.quota_value
+
+    return res
+
+
+@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
+       stop_max_attempt_number=50)
+@utils.no_4byte_params
 def create_lock(context, lock_key, session):
     """Try to create lock record."""
-    try:
-        lock = models.ArtifactLock()
-        lock.id = lock_key
-        lock.save(session=session)
-        return lock.id
-    except (sqlalchemy.exc.IntegrityError, db_exception.DBDuplicateEntry):
-        msg = _("Cannot lock an item with key %s. "
-                "Lock already acquired by other request") % lock_key
-        raise exception.Conflict(msg)
+    with session.begin():
+        existing = session.query(models.ArtifactLock).get(lock_key)
+        if existing is None:
+            try:
+                lock = models.ArtifactLock()
+                lock.id = lock_key
+                lock.save(session=session)
+                return lock.id
+            except (sqlalchemy.exc.IntegrityError,
+                    db_exception.DBDuplicateEntry):
+                msg = _("Cannot lock an item with key %s. "
+                        "Lock already acquired by other request") % lock_key
+                raise exception.Conflict(msg)
+        else:
+            if timeutils.is_older_than(existing.acquired_at, 5):
+                existing.acquired_at = timeutils.utcnow()
+                existing.save(session)
+                return existing.id
+            else:
+                msg = _("Cannot lock an item with key %s. "
+                        "Lock already acquired by other request") % lock_key
+                raise exception.Conflict(msg)
 
 
 @retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
        stop_max_attempt_number=50)
 def delete_lock(context, lock_id, session):
-    try:
+    with session.begin():
         session.query(models.ArtifactLock).filter_by(id=lock_id).delete()
+
+
+@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
+       stop_max_attempt_number=50)
+def save_blob_data(context, blob_data_id, data, session):
+    """Save blob data to database."""
+    with session.begin():
+        blob_data = models.ArtifactBlobData()
+        blob_data.id = blob_data_id
+        blob_data.data = data.read()
+        blob_data.save(session=session)
+        return "sql://" + blob_data.id
+
+
+@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
+       stop_max_attempt_number=50)
+def save_blob_data_batch(context, blobs, session):
+    """Perform batch uploading to database."""
+    with session.begin():
+
+        locations = []
+
+        # blobs is a list of tuples (blob_data_id, data)
+        for blob_data_id, data in blobs:
+            blob_data = models.ArtifactBlobData()
+            blob_data.id = blob_data_id
+            blob_data.data = data.read()
+            session.add(blob_data)
+            locations.append("sql://" + blob_data.id)
+
+        session.flush()
+
+    return locations
+
+
+@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
+       stop_max_attempt_number=50)
+def get_blob_data(context, uri, session):
+    """Download blob data from database."""
+
+    blob_data_id = uri[6:]
+    try:
+        blob_data = session.query(
+            models.ArtifactBlobData).filter_by(id=blob_data_id).one()
     except orm.exc.NoResultFound:
-        msg = _("Cannot delete a lock with id %s.") % lock_id
+        msg = _("Cannot find a blob data with id %s.") % blob_data_id
         raise exception.NotFound(msg)
+    return blob_data.data
+
+
+@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
+       stop_max_attempt_number=50)
+def delete_blob_data(context, uri, session):
+    """Delete blob data from database."""
+    with session.begin():
+        blob_data_id = uri[6:]
+        session.query(
+            models.ArtifactBlobData).filter_by(id=blob_data_id).delete()
